# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Search engine API."""

import hashlib
from functools import partial

import six
from elasticsearch import VERSION as ES_VERSION
from elasticsearch_dsl import FacetedSearch, Search
from elasticsearch_dsl.faceted_search import FacetedResponse
from elasticsearch_dsl.query import Bool, Ids
from flask import current_app, request

from .proxies import current_search_client
from .utils import build_alias_name


class DefaultFilter(object):
    """Shortcut for defining default filters with query parser."""

    def __init__(self, query=None, query_parser=None):
        """Build filter property with query parser."""
        self._query = query
        self.query_parser = query_parser or (lambda x: x)

    @property
    def query(self):
        """Build lazy query if needed."""
        return self._query() if callable(self._query) else self._query

    def __get__(self, obj, objtype):
        """Return parsed query."""
        return self.query_parser(self.query)


class MinShouldMatch(str):
    """Work-around for Elasticsearch DSL problem.

    The ElasticSearch DSL Bool query tries to inspect the
    ``minimum_should_match`` parameter, but understands only integers and not
    queries like "0<1". This class circumvents the specific problematic clause
    in Elasticsearch DSL.
    """

    def __lt__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False

    def __le__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False

    def __gt__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False

    def __ge__(self, other):
        """Circumvent problematic Elasticsearch DSL clause."""
        return False


class BaseRecordsSearch(Search):
    """Example subclass for searching records using Elastic DSL."""

    class Meta:
        """Configuration for ``Search`` and ``FacetedSearch`` classes."""

        index = '*'
        doc_types = None
        fields = ('*', )
        facets = {}

        default_filter = None
        """Default filter added to search body.

        Example: ``default_filter = DefaultFilter('_access.owner:"1"')``.
        """

    def __init__(self, **kwargs):
        """Use Meta to set kwargs defaults."""
        kwargs.setdefault('index', getattr(self.Meta, 'index', None))
        kwargs.setdefault('doc_type', getattr(self.Meta, 'doc_types', None))
        kwargs.setdefault('using', current_search_client)
        kwargs.setdefault('extra', {})

        min_score = current_app.config.get('SEARCH_RESULTS_MIN_SCORE')
        if min_score:
            kwargs['extra'].update(min_score=min_score)

        super(BaseRecordsSearch, self).__init__(**kwargs)

        default_filter = getattr(self.Meta, 'default_filter', None)
        if default_filter:
            # NOTE: https://github.com/elastic/elasticsearch/issues/21844
            self.query = Bool(minimum_should_match=MinShouldMatch("0<1"),
                              filter=default_filter)

    def get_record(self, id_):
        """Return a record by its identifier.

        :param id_: The record identifier.
        :returns: The record.
        """
        return self.query(Ids(values=[str(id_)]))

    def get_records(self, ids):
        """Return records by their identifiers.

        :param ids: A list of record identifier.
        :returns: A list of records.
        """
        return self.query(Ids(values=[str(id_) for id_ in ids]))

    @classmethod
    def faceted_search(cls, query=None, filters=None, search=None):
        """Return faceted search instance with defaults set.

        :param query: Elastic DSL query object (``Q``).
        :param filters: Dictionary with selected facet values.
        :param search: An instance of ``Search`` class. (default: ``cls()``).
        """
        search_ = search or cls()

        class RecordsFacetedSearch(FacetedSearch):
            """Pass defaults from ``cls.Meta`` object."""

            index = build_alias_name(search_._index[0])
            doc_types = getattr(search_.Meta, 'doc_types', ['*'])
            fields = getattr(search_.Meta, 'fields', ('*', ))
            facets = getattr(search_.Meta, 'facets', {})

            def search(self):
                """Use ``search`` or ``cls()`` instead of default Search."""
                # Later versions of `elasticsearch-dsl` (>=5.1.0) changed the
                # Elasticsearch FacetedResponse class constructor signature.

                if ES_VERSION[0] > 2:
                    return search_.response_class(FacetedResponse)
                return search_.response_class(partial(FacetedResponse, self))

        return RecordsFacetedSearch(query=query, filters=filters or {})

    def with_preference_param(self):
        """Add the preference param to the ES request and return a new Search.

        The preference param avoids the bouncing effect with multiple
        replicas, documented on ES documentation.
        See: https://www.elastic.co/guide/en/elasticsearch/guide/current
        /_search_options.html#_preference for more information.
        """
        user_hash = self._get_user_hash()
        if user_hash:
            return self.params(preference=user_hash)
        return self

    def _get_user_agent(self):
        """Retrieve the request's User-Agent, if available.

        Taken from Flask Login utils.py.
        """
        user_agent = request.headers.get('User-Agent')
        if user_agent:
            user_agent = user_agent.encode('utf-8')
        return user_agent or ''

    def _get_user_hash(self):
        """Calculate a digest based on request's User-Agent and IP address."""
        if request:
            user_hash = '{ip}-{ua}'.format(ip=request.remote_addr,
                                           ua=self._get_user_agent())
            alg = hashlib.md5()
            alg.update(user_hash.encode('utf8'))
            return alg.hexdigest()
        return None


class PrefixedIndexList(list):
    """Custom list type for avoiding double prefixing."""

    pass


class RecordsSearch(BaseRecordsSearch):
    """Example subclass for searching records using index prefixing."""

    def __init__(self, **kwargs):
        """Using PrefixedIndexList type to avoid double prefixing."""
        # at object instantiation, kwargs['index'] is not defined.
        # Elasticsearch-dsl-py re-instantiated the object at each search
        # by cloning it and passing as kwargs the list of indices
        # kwargs['index'] = ['index-name1', 'index-name2']
        _index_param = kwargs.get('index', getattr(self.Meta, 'index', None))
        if not isinstance(_index_param, PrefixedIndexList):
            if isinstance(_index_param, (tuple, list)):
                _prefixed_index_list = [
                    build_alias_name(_index)
                    for _index in _index_param
                ]
                kwargs.update({'index': _prefixed_index_list})
            elif isinstance(_index_param, six.string_types):
                _splitted_index = _index_param.strip().split(',')
                if len(_splitted_index) > 1:
                    _prefix_index_list = [
                        build_alias_name(_index)
                        for _index in _splitted_index]
                    _prefix_index_param = ','.join(_prefix_index_list)
                    kwargs.update({'index': _prefix_index_param})
                else:
                    kwargs.update({'index': build_alias_name(_index_param)})
                _index_param = [_index_param]
            self._original_index = _index_param

        super(RecordsSearch, self).__init__(**kwargs)
        if self._index:
            self._index = PrefixedIndexList(self._index)

    def _clone(self):
        """Clone `_original_index` attribute.

        During re-instantiation Elasticsearch-dsl-py calls `self._clone`
        to copy over the search object. We override the method so we can
        copy the `_original_index` attribute.
        """
        s = super(RecordsSearch, self)._clone()
        s._original_index = self._original_index
        return s


UnPrefixedRecordsSearch = BaseRecordsSearch



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os
import shutil
import sys
import tempfile

import pytest
from flask import Flask
from invenio_db import InvenioDB
from sqlalchemy_utils.functions import create_database, database_exists, \
    drop_database

from invenio_search import InvenioSearch

sys.path.append(
    os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
                 'tests/mock_module'))


@pytest.fixture()
def app():
    """Flask application fixture."""
    # Set temporary instance path for sqlite
    instance_path = tempfile.mkdtemp()
    app = Flask('testapp', instance_path=instance_path)
    app.config.update(
        TESTING=True
    )
    InvenioSearch(app)

    with app.app_context():
        yield app

    # Teardown instance path.
    shutil.rmtree(instance_path)


def mock_iter_entry_points_factory(data, mocked_group):
    """Create a mock iter_entry_points function."""
    from pkg_resources import iter_entry_points

    def entrypoints(group, name=None):
        if group == mocked_group:
            for entrypoint in data:
                yield entrypoint
        else:
            for x in iter_entry_points(group=group, name=name):
                yield x
    return entrypoints


@pytest.fixture()
def template_entrypoints():
    """Declare some events by mocking the invenio_stats.events entrypoint.

    It yields a list like [{event_type: <event_type_name>}, ...].
    """
    eps = []
    for idx in range(5):
        event_type_name = 'mock_module'
        from pkg_resources import EntryPoint
        entrypoint = EntryPoint(event_type_name, event_type_name)
        entrypoint.load = lambda: lambda: ['mock_module.templates']
        eps.append(entrypoint)

    entrypoints = mock_iter_entry_points_factory(
        eps, 'invenio_search.templates')
    return entrypoints



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

import pytest
from elasticsearch import VERSION as ES_VERSION
from mock import patch

from invenio_search.utils import build_index_name, schema_to_index


@pytest.mark.parametrize(
    ('schema, expected, index_names'),
    (
        (
            'records/record-v1.0.0.json',
            ('records-record-v1.0.0', 'record-v1.0.0'),
            None,
        ),
        (
            '/records/record-v1.0.0.json',
            ('records-record-v1.0.0', 'record-v1.0.0'),
            None,
        ),
        (
            'default-v1.0.0.json',
            ('default-v1.0.0', 'default-v1.0.0'),
            None,
        ),
        (
            'default-v1.0.0.json',
            (None, None),
            [],
        ),
        (
            'invalidextension',
            (None, None),
            None,
        ),
    ),
)
def test_schema_to_index(schema, expected, index_names, app):
    """Test the expected value of schema to index."""
    result = schema_to_index(schema, index_names=index_names)
    if ES_VERSION[0] >= 7 and expected[0]:
        expected = (expected[0], '_doc')
    assert result == expected


@pytest.mark.parametrize(('parts', 'prefix', 'suffix', 'expected'), [
    (['records'], '', '', 'records'),
    (['records'], 'foo-', '', 'foo-records'),
    (['records', 'record'], 'foo-', '', 'foo-records-record'),
    (['records', 'record'], '', '-new', 'records-record-new'),
    (['test', 'recs', 'rec'], 'foo-', '-old', 'foo-test-recs-rec-old'),
])
def test_build_suffix_index_name(app, parts, prefix, suffix, expected):
    app.config.update(SEARCH_INDEX_PREFIX=prefix)
    assert build_index_name(parts, suffix=suffix, app=app) == expected


def test_schema_to_index_with_names(app):
    """Test that prefix is added to the index when creating it."""
    result = schema_to_index(
        'default-v1.0.0.json',
        index_names=['default-v1.0.0']
    )
    doc_type = '_doc' if ES_VERSION[0] >= 7 else 'default-v1.0.0'
    assert result == ('default-v1.0.0', doc_type)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for information retrieval."""

from __future__ import absolute_import, print_function

import errno
import json
import os
import warnings

from elasticsearch import VERSION as ES_VERSION
from elasticsearch import Elasticsearch
from pkg_resources import iter_entry_points, resource_filename, \
    resource_isdir, resource_listdir
from werkzeug.utils import cached_property

from . import config
from .cli import index as index_cmd
from .errors import IndexAlreadyExistsError
from .utils import build_alias_name, build_index_from_parts, \
    build_index_name, timestamp_suffix


class _SearchState(object):
    """Store connection to elastic client and registered indexes."""

    def __init__(self, app,
                 entry_point_group_mappings=None,
                 entry_point_group_templates=None,
                 **kwargs):
        """Initialize state.

        :param app: An instance of :class:`~flask.app.Flask`.
        :param entry_point_group_mappings:
            The entrypoint group name to load mappings.
        :param entry_point_group_templates:
            The entrypoint group name to load templates.
        """
        self.app = app
        self.mappings = {}
        self.aliases = {}
        self._client = kwargs.get('client')
        self.entry_point_group_templates = entry_point_group_templates
        self._current_suffix = None

        if entry_point_group_mappings:
            self.load_entry_point_group_mappings(entry_point_group_mappings)

        if ES_VERSION[0] in (2, 5):
            warnings.warn(
                "Elasticsearch v2 and v5 support will be removed.",
                DeprecationWarning)

    @property
    def current_suffix(self):
        """Return the current suffix."""
        if self._current_suffix is None:
            self._current_suffix = timestamp_suffix()
        return self._current_suffix

    @cached_property
    def templates(self):
        """Generate a dictionary with template names and file paths."""
        templates = {}
        result = []
        if self.entry_point_group_templates:
            result = self.load_entry_point_group_templates(
                self.entry_point_group_templates) or []

        for template in result:
            for name, path in template.items():
                templates[name] = path

        return templates

    def register_mappings(self, alias, package_name):
        """Register mappings from a package under given alias.

        :param alias: The alias.
        :param package_name: The package name.
        """
        # For backwards compatibility, we also allow for ES2 mappings to be
        # placed at the root level of the specified package path, and not in
        # the `<package-path>/v2` directory.
        if ES_VERSION[0] == 2:
            try:
                resource_listdir(package_name, 'v2')
                package_name += '.v2'
            except (OSError, IOError) as ex:
                if getattr(ex, 'errno', 0) != errno.ENOENT:
                    raise
                warnings.warn(
                    "Having mappings in a path which doesn't specify the "
                    "Elasticsearch version is deprecated. Please move your "
                    "mappings to a subfolder named according to the "
                    "Elasticsearch version which your mappings are intended "
                    "for. (e.g. '{}/v2/{}')".format(
                        package_name, alias),
                    PendingDeprecationWarning)
        else:
            package_name = '{}.v{}'.format(package_name, ES_VERSION[0])

        def _walk_dir(aliases, *parts):
            root_name = build_index_from_parts(*parts)
            resource_name = os.path.join(*parts)

            data = aliases.get(root_name, {})

            for filename in resource_listdir(package_name, resource_name):
                file_path = os.path.join(resource_name, filename)

                if resource_isdir(package_name, file_path):
                    _walk_dir(data, *(parts + (filename, )))
                    continue

                filename_root, ext = os.path.splitext(filename)
                if ext not in {'.json', }:
                    continue

                index_name = build_index_from_parts(
                    *(parts + (filename_root, ))
                )
                assert index_name not in data, 'Duplicate index'
                filename = resource_filename(
                    package_name, os.path.join(resource_name, filename))
                data[index_name] = filename
                self.mappings[index_name] = filename

            aliases[root_name] = data

        # Start the recursion here:
        _walk_dir(self.aliases, alias)

    def register_templates(self, module):
        """Register templates from the provided module.

        :param module: The templates module.
        """
        try:
            resource_listdir(module, 'v{}'.format(ES_VERSION[0]))
            module = '{}.v{}'.format(module, ES_VERSION[0])
        except (OSError, IOError) as ex:
            if getattr(ex, 'errno', 0) == errno.ENOENT:
                raise OSError(
                    "Please move your templates to a subfolder named "
                    "according to the Elasticsearch version "
                    "which your templates are intended "
                    "for. (e.g. '{}')".format(version_module))
        result = {}

        def _walk_dir(*parts):
            parts = parts or tuple()
            resource_name = os.path.join(*parts) if parts else ''

            for filename in resource_listdir(module, resource_name):
                file_path = os.path.join(resource_name, filename)

                if resource_isdir(module, file_path):
                    _walk_dir(*(parts + (filename, )))
                    continue

                filename_root, ext = os.path.splitext(filename)
                if ext not in {'.json', }:
                    continue

                template_name = build_index_from_parts(
                    *(parts + (filename_root, ))
                )
                result[template_name] = resource_filename(
                    module, os.path.join(resource_name, filename))

        # Start the recursion here:
        _walk_dir()
        return result

    def load_entry_point_group_mappings(self, entry_point_group_mappings):
        """Load actions from an entry point group."""
        for ep in iter_entry_points(group=entry_point_group_mappings):
            self.register_mappings(ep.name, ep.module_name)

    def load_entry_point_group_templates(self, entry_point_group_templates):
        """Load actions from an entry point group."""
        result = []
        for ep in iter_entry_points(group=entry_point_group_templates):
            with self.app.app_context():
                for template_dir in ep.load()():
                    result.append(self.register_templates(template_dir))
        return result

    def _client_builder(self):
        """Build Elasticsearch client."""
        client_config = self.app.config.get('SEARCH_CLIENT_CONFIG') or {}
        client_config.setdefault(
            'hosts', self.app.config.get('SEARCH_ELASTIC_HOSTS'))
        return Elasticsearch(**client_config)

    @property
    def client(self):
        """Return client for current application."""
        if self._client is None:
            self._client = self._client_builder()
        return self._client

    def flush_and_refresh(self, index):
        """Flush and refresh one or more indices.

        .. warning::

           Do not call this method unless you know what you are doing. This
           method is only intended to be called during tests.
        """
        prefixed_index = build_alias_name(index, app=self.app)
        self.client.indices.flush(wait_if_ongoing=True, index=prefixed_index)
        self.client.indices.refresh(index=prefixed_index)
        self.client.cluster.health(
            wait_for_status='yellow', request_timeout=30)
        return True

    @property
    def cluster_version(self):
        """Get version of Elasticsearch running on the cluster."""
        versionstr = self.client.info()['version']['number']
        return [int(x) for x in versionstr.split('.')]

    @property
    def active_aliases(self):
        """Get a filtered list of aliases based on configuration.

        Returns aliases and their mappings that are defined in the
        `SEARCH_MAPPINGS` config variable. If the `SEARCH_MAPPINGS` is set to
        `None` (the default), all aliases are included.
        """
        whitelisted_aliases = self.app.config.get('SEARCH_MAPPINGS')
        if whitelisted_aliases is None:
            return self.aliases
        else:
            return {k: v for k, v in self.aliases.items()
                    if k in whitelisted_aliases}

    def _get_indices(self, tree_or_filename):
        for name, value in tree_or_filename.items():
            if isinstance(value, dict):
                for result in self._get_indices(value):
                    yield result
            else:
                yield name

    def create_index(self, index, mapping_path=None, prefix=None, suffix=None,
                     create_write_alias=True, ignore=None, dry_run=False):
        """Create index with a write alias."""
        mapping_path = mapping_path or self.mappings[index]

        final_alias = None
        final_index = None
        index_result = None, None
        alias_result = None, None
        # To prevent index init --force from creating a suffixed
        # index if the current instance is running without suffixes
        # make sure there is no index with the same name as the
        # alias name (i.e. the index name without the suffix).
        with open(mapping_path, 'r') as body:
            final_index = build_index_name(
                index, prefix=prefix, suffix=suffix, app=self.app)
            if create_write_alias:
                final_alias = build_alias_name(
                    index, prefix=prefix, app=self.app)
            index_result = (
                final_index,
                self.client.indices.create(
                    index=final_index,
                    body=json.load(body),
                    ignore=ignore,
                ) if not dry_run else None
            )
            if create_write_alias:
                alias_result = (
                    final_alias,
                    self.client.indices.put_alias(
                        index=final_index,
                        name=final_alias,
                        ignore=ignore,
                    ) if not dry_run else None
                )
        return index_result, alias_result

    def create(self, ignore=None, ignore_existing=False, index_list=None):
        """Yield tuple with created index name and responses from a client."""
        ignore = ignore or []
        new_indices = {}
        actions = []
        if ignore_existing and not ignore:
            ignore = [400]
        elif ignore_existing and 400 not in ignore:
            ignore.append(400)

        def ensure_not_exists(name):
            if not ignore_existing and self.client.indices.exists(name):
                raise IndexAlreadyExistsError(
                    'index/alias with name "{}" already exists'.format(name))

        def _build(tree_or_filename, alias=None):
            """Build a list of index/alias actions to perform."""
            for name, value in tree_or_filename.items():
                if isinstance(value, dict):
                    _build(value, alias=name)
                else:
                    if index_list and name not in index_list:
                        continue
                    index_result, alias_result = \
                        self.create_index(
                            name,
                            ignore=ignore,
                            dry_run=True
                        )
                    ensure_not_exists(index_result[0])
                    new_indices[name] = index_result[0]
                    if alias_result[0]:
                        ensure_not_exists(alias_result[0])
                        actions.append(dict(
                            type='create_index',
                            index=name,
                            create_write_alias=True
                        ))
                    else:
                        actions.append(dict(
                            type='create_index',
                            index=name,
                            create_write_alias=False
                        ))
            if alias:
                alias_indices = self._get_indices(tree_or_filename)
                alias_indices = [
                    new_indices[i] for i in alias_indices if i in new_indices
                ]
                if alias_indices:
                    alias_name = build_alias_name(alias, app=self.app)
                    ensure_not_exists(alias_name)
                    actions.append(dict(
                        type='create_alias',
                        index=alias_indices,
                        alias=alias_name
                    ))

        _build(self.active_aliases)

        for action in actions:
            if action['type'] == 'create_index':
                index_result, alias_result = self.create_index(
                    action['index'],
                    create_write_alias=action.get('create_write_alias', True),
                    ignore=ignore
                )
                yield index_result
                if alias_result[0]:
                    yield alias_result
            elif action['type'] == 'create_alias':
                yield action['alias'], self.client.indices.put_alias(
                    index=action['index'],
                    name=action['alias'],
                    ignore=ignore,
                )

    def put_templates(self, ignore=None):
        """Yield tuple with registered template and response from client."""
        ignore = ignore or []

        def _replace_prefix(template_path, body):
            """Replace index prefix in template request body."""
            pattern = '__SEARCH_INDEX_PREFIX__'

            prefix = self.app.config['SEARCH_INDEX_PREFIX'] or ''
            if prefix:
                assert pattern in body, "You are using the prefix `{0}`, "
                "but the template `{1}` does not contain the "
                "pattern `{2}`.".format(prefix, template_path, pattern)

            return body.replace(pattern, prefix)

        def _put_template(template):
            """Put template in search client."""
            with open(self.templates[template], 'r') as fp:
                body = fp.read()
                replaced_body = _replace_prefix(self.templates[template], body)
                template_name = build_alias_name(template, app=self.app)
                return self.templates[template],\
                    self.client.indices.put_template(
                        name=template_name,
                        body=json.loads(replaced_body),
                        ignore=ignore,
                )

        for template in self.templates:
            yield _put_template(template)

    def delete(self, ignore=None, index_list=None):
        """Yield tuple with deleted index name and responses from a client."""
        ignore = ignore or []

        def _delete(tree_or_filename, alias=None):
            """Delete indexes and aliases by walking DFS."""
            # Iterate over aliases:
            for name, value in tree_or_filename.items():
                if isinstance(value, dict):
                    for result in _delete(value, alias=name):
                        yield result
                else:
                    if index_list and name not in index_list:
                        continue
                    # Resolve values to suffixed (or not) indices
                    prefixed_index = build_alias_name(name, app=self.app)
                    lookup_response = self.client.indices.get_alias(
                        index=prefixed_index, ignore=[404])
                    if 'error' in lookup_response:
                        indices_to_delete = []
                    else:
                        indices_to_delete = list(lookup_response.keys())
                    if len(indices_to_delete) == 0:
                        pass
                    elif len(indices_to_delete) == 1:
                        yield name, self.client.indices.delete(
                            index=indices_to_delete[0],
                            ignore=ignore,
                        )
                    else:
                        warnings.warn((
                            'Multiple indices found during deletion of '
                            '{name}: {indices}. Deletion was skipped for them.'
                        ).format(name=name, indices=indices_to_delete))

        for result in _delete(self.active_aliases):
            yield result


class InvenioSearch(object):
    """Invenio-Search extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        self._clients = {}

        if app:
            self.init_app(app, **kwargs)

    def init_app(self, app,
                 entry_point_group_mappings='invenio_search.mappings',
                 entry_point_group_templates='invenio_search.templates',
                 **kwargs):
        """Flask application initialization.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        self.init_config(app)

        app.cli.add_command(index_cmd)

        state = _SearchState(
            app,
            entry_point_group_mappings=entry_point_group_mappings,
            entry_point_group_templates=entry_point_group_templates,
            **kwargs
        )
        self._state = app.extensions['invenio-search'] = state

    @staticmethod
    def init_config(app):
        """Initialize configuration.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        for k in dir(config):
            if k.startswith('SEARCH_'):
                app.config.setdefault(k, getattr(config, k))

    def __getattr__(self, name):
        """Proxy to state object."""
        return getattr(self._state, name, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import hashlib

from elasticsearch import VERSION as ES_VERSION
from elasticsearch_dsl import Q, Search
from flask import request

from invenio_search.api import BaseRecordsSearch, DefaultFilter, RecordsSearch


def test_empty_query(app):
    """Test building an empty query."""
    q = RecordsSearch()
    if ES_VERSION[0] >= 7:
        q.to_dict() == {}
    else:
        q.to_dict() == {'query': {'match_all': {}}}

    q = RecordsSearch.faceted_search('')
    if ES_VERSION[0] >= 7:
        q._s.to_dict() == {'highlight': {'fields': {'*': {}}}}
    else:
        q._s.to_dict() == {'query': {'match_all': {}}}

    q = RecordsSearch()[10]
    assert q.to_dict()['from'] == 10
    assert q.to_dict()['size'] == 1

    q = q[10:20]
    assert q.to_dict()['from'] == 10
    assert q.to_dict()['size'] == 10

    q = q.sort({'field1': {'order': 'asc'}})
    assert q.to_dict()['sort'][0] == {'field1': {'order': 'asc'}}

    q = q.sort()
    assert 'sort' not in q.to_dict()

    q = q.sort('-field1')
    assert q.to_dict()['sort'][0] == {'field1': {'order': 'desc'}}

    q = q.sort('field2', {'field3': {'order': 'asc'}})
    assert q.to_dict()['sort'][0] == 'field2'
    assert q.to_dict()['sort'][1] == {'field3': {'order': 'asc'}}
    q.sort()

    q = RecordsSearch()
    q = q.highlight('field1', index_options='offsets')
    assert len(q.to_dict()['highlight']['fields']) == 1
    assert q.to_dict()['highlight']['fields']['field1'] == {
        'index_options': 'offsets'
    }

    q = q.highlight('field2')
    assert len(q.to_dict()['highlight']['fields']) == 2
    assert q.to_dict()['highlight']['fields']['field1'] == {
        'index_options': 'offsets'
    }
    assert q.to_dict()['highlight']['fields']['field2'] == {}

    q = q.highlight()
    assert 'highligth' not in q.to_dict()


def test_elasticsearch_query(app):
    """Test building a real query."""
    from flask import g

    class TestSearch(RecordsSearch):
        class Meta:
            default_filter = DefaultFilter(
                lambda: Q('terms', public=g.public)
            )

    g.public = 1
    q = TestSearch()
    assert q.to_dict()['query'] == {
        'bool': {
            'minimum_should_match': "0<1",
            'filter': [{'terms': {'public': 1}}]
        }
    }
    g.public = 0
    q = TestSearch()
    q = q.query(Q('match', title='Higgs'))
    assert q.to_dict()['query']['bool']['filter'] == [
        {'terms': {'public': 0}}
    ]
    assert q.to_dict()['query']['bool']['must'] == [
        {'match': {'title': 'Higgs'}}
    ]


class SpySearch(Search):
    exposed_params = {}

    def params(self, **kwargs):
        self.exposed_params.update(kwargs)
        return super(SpySearch, self).params(**kwargs)


def test_es_preference_param_no_request(app):
    """Test that the preference param is not added when not in a request."""
    BaseRecordsSearch.__bases__ = (SpySearch,)

    rs = RecordsSearch()
    new_rs = rs.with_preference_param()
    assert new_rs.exposed_params == {}


def test_es_preference_param(app):
    """Test the preference param is correctly added in a request."""
    BaseRecordsSearch.__bases__ = (SpySearch,)

    with app.test_request_context('/', headers={'User-Agent': 'Chrome'},
                                  environ_base={'REMOTE_ADDR': '212.54.1.8'}):
        rs = RecordsSearch()
        new_rs = rs.with_preference_param()

        alg = hashlib.md5()
        encoded_user_agent = 'Chrome'.encode('utf8')
        encoded_user_string = '{ip}-{ua}'.format(ip=request.remote_addr,
                                                 ua=encoded_user_agent)
        alg.update(encoded_user_string.encode('utf8'))
        digest = alg.hexdigest()

        assert new_rs.exposed_params == dict(preference=digest)


def test_elasticsearch_query_min_score(app):
    """Test building a query with min_score."""
    app.config.update(SEARCH_RESULTS_MIN_SCORE=0.1)

    q = RecordsSearch()
    q = q.query(Q('match', title='Higgs'))

    search_dict = q.to_dict()
    assert 'min_score' in search_dict
    assert search_dict['min_score'] == app.config['SEARCH_RESULTS_MIN_SCORE']


def _test_original_index_is_stored_when_prefixing(q, prefixed_index,
                                                  original_index):
    """Test original index is stored."""

    q = q.query(Q('match', title='Higgs'))
    q = q.sort()
    search_dict = q.to_dict()
    assert 'query' in search_dict
    assert q._index == prefixed_index
    assert q._original_index == original_index


def test_prefix_index_from_meta(app):
    """Test that index is prefixed when you use `Meta.index` field."""
    class TestSearch(RecordsSearch):
        class Meta:
            index = 'myindex'

    prefix_value = 'myprefix-'
    index_value = TestSearch.Meta.index
    app.config['SEARCH_INDEX_PREFIX'] = prefix_value

    prefixed_index = ['{}{}'.format(prefix_value, index_value)]
    q = TestSearch()
    _test_original_index_is_stored_when_prefixing(q, prefixed_index,
                                                  [index_value])


def test_prefix_index_from_kwargs(app):
    """Test that index is prefixed when pass it through kwargs."""
    prefix_value = 'myprefix-'
    index_value = 'myindex'
    app.config['SEARCH_INDEX_PREFIX'] = prefix_value

    prefixed_index = ['{}{}'.format(prefix_value, index_value)]
    q = RecordsSearch(index=index_value)
    _test_original_index_is_stored_when_prefixing(q, prefixed_index,
                                                  [index_value])


def test_prefix_index_list(app):
    """Test that index is prefixed when pass it through kwargs."""
    prefix_value = 'myprefix-'
    index_value = ['myindex', 'myanotherindex']
    app.config['SEARCH_INDEX_PREFIX'] = prefix_value

    prefixed_index = ['{}{}'.format(prefix_value, _index)
                      for _index in index_value]

    q = RecordsSearch(index=index_value)
    _test_original_index_is_stored_when_prefixing(q, prefixed_index,
                                                  index_value)


def test_prefix_multi_index_string(app):
    """Test that index is prefixed when pass it through kwargs."""
    prefix_value = 'myprefix-'
    index_value = 'myindex,myanotherindex'
    app.config['SEARCH_INDEX_PREFIX'] = prefix_value

    prefixed_index = [','.join(['{}{}'.format(prefix_value, _index)
                                for _index in index_value.split(',')])]
    q = RecordsSearch(index=index_value)
    _test_original_index_is_stored_when_prefixing(q, prefixed_index,
                                                  [index_value])



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Configuration options for Invenio-Search.

The documentation for the configuration is in docs/configuration.rst.
"""

#
# ELASTIC configuration
#

SEARCH_CLIENT_CONFIG = None
"""Dictionary of options for the Elasticsearch client.

The value of this variable is passed to :py:class:`elasticsearch.Elasticsearch`
as keyword arguments and is used to configure the client. See the available
keyword arguments in the two following classes:

- :py:class:`elasticsearch.Elasticsearch`
- :py:class:`elasticsearch.Transport`

If you specify the key ``hosts`` in this dictionary, the configuration variable
:py:class:`~invenio_search.config.SEARCH_ELASTIC_HOSTS` will have no effect.
"""

SEARCH_ELASTIC_HOSTS = None  # default localhost
"""Elasticsearch hosts.

By default, Invenio connects to ``localhost:9200``.

The value of this variable is a list of dictionaries, where each dictionary
represents a host. The available keys in each dictionary is determined by the
connection class:

- :py:class:`elasticsearch.connection.Urllib3HttpConnection` (default)
- :py:class:`elasticsearch.connection.RequestsHttpConnection`

You can change the connection class via the
:py:class:`~invenio_search.config.SEARCH_CLIENT_CONFIG`. If you specified the
``hosts`` key in :py:class:`~invenio_search.config.SEARCH_CLIENT_CONFIG` then
this configuration variable will have no effect.
"""


SEARCH_MAPPINGS = None  # loads all mappings and creates aliases for them
"""List of aliases for which, their search mappings should be created.

- If `None` all aliases (and their search mappings) defined through the
  ``invenio_search.mappings`` entry point in setup.py will be created.
- Provide an empty list ``[]`` if no aliases (or their search mappings)
  should be created.

For example if you don't want to create aliases
and their mappings for `authors`:

.. code-block:: python

    # in your `setup.py` you would specify:
    entry_points={
        'invenio_search.mappings': [
            'records = invenio_foo_bar.mappings',
            'authors = invenio_foo_bar.mappings',
        ],
    }

    # and in your config.py
    SEARCH_MAPPINGS = ['records']
"""

SEARCH_RESULTS_MIN_SCORE = None
"""If set, the `min_score` parameter is added to each search request body.

The `min_score` parameter excludes results which have a `_score` less than
the minimum specified in `min_score`.

Note that the `max_score` varies depending on the number of results for a given
search query and it is not absolute value. Therefore, setting `min_score` too
high can lead to 0 results because it can be higher than any result's `_score`.

Please refer to `Elasticsearch min_score documentation
<https://www.elastic.co/guide/en/elasticsearch/reference/current/
search-request-min-score.html>`_ for more information.
"""

SEARCH_INDEX_PREFIX = ''
"""Any index, alias and templates will be prefixed with this string.

Useful to host multiple instances of the app on the same Elasticsearch cluster,
for example on one app you can set it to `dev-` and on the other to `prod-`,
and each will create non-colliding indices prefixed with the corresponding
string.

Usage example:

.. code-block:: python

    # in your config.py
    SEARCH_INDEX_PREFIX = 'prod-'

For templates, ensure that the prefix `__SEARCH_INDEX_PREFIX__` is added to
your index names. This pattern will be replaced by the prefix config value.

Usage example in your template.json:

.. code-block:: json

    {
        "index_patterns": ["__SEARCH_INDEX_PREFIX__myindex-name-*"]
    }
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Test CLI."""

from __future__ import absolute_import, print_function

import ast

import pytest
from click.testing import CliRunner
from elasticsearch import VERSION as ES_VERSION
from elasticsearch.exceptions import NotFoundError
from flask.cli import ScriptInfo
from mock import patch

from invenio_search.cli import index as cmd
from invenio_search.proxies import current_search_client


def test_init(app, template_entrypoints):
    """Run client initialization."""
    suffix = '-abc'
    search = app.extensions['invenio-search']
    search._current_suffix = suffix
    search.register_mappings('records', 'mock_module.mappings')

    assert 'records' in search.aliases
    assert set(search.aliases['records']) == {
        'records-authorities',
        'records-bibliographic',
        'records-default-v1.0.0'
    }
    assert set(search.aliases['records']['records-authorities']) == {
        'records-authorities-authority-v1.0.0',
    }
    assert set(search.aliases['records']['records-bibliographic']) == {
        'records-bibliographic-bibliographic-v1.0.0',
    }
    assert set(search.mappings.keys()) == {
        'records-authorities-authority-v1.0.0',
        'records-bibliographic-bibliographic-v1.0.0',
        'records-default-v1.0.0'
    }
    assert 3 == len(search.mappings)

    with patch('invenio_search.ext.iter_entry_points',
               return_value=template_entrypoints('invenio_search.templates')):
        if ES_VERSION[0] == 2:
            assert len(search.templates.keys()) == 2
            assert 'record-view-v1' in search.templates
            assert 'subdirectory-file-download-v1' in search.templates
        else:
            assert len(search.templates.keys()) == 1
            assert 'record-view-v{}'.format(ES_VERSION[0]) in search.templates

    current_search_client.indices.delete_alias('_all', '_all', ignore=[400,
                                                                       404])
    current_search_client.indices.delete('*')
    aliases = current_search_client.indices.get_alias()
    assert 0 == len(aliases)

    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda info: app)

    with runner.isolated_filesystem():
        result = runner.invoke(cmd, ['init', '--force'], obj=script_info)
        assert result.exit_code == 0
        if ES_VERSION[0] == 2:
            assert current_search_client.indices.exists_template(
                'subdirectory-file-download-v1')
            assert current_search_client.indices.exists_template(
                'record-view-v1')
        else:
            assert current_search_client.indices.exists_template(
                'record-view-v{}'.format(ES_VERSION[0]))
        assert 0 == result.exit_code

    aliases = current_search_client.indices.get_alias()
    assert 8 == sum(len(idx.get('aliases', {})) for idx in aliases.values())

    assert current_search_client.indices.exists(list(search.mappings.keys()))

    # Clean-up:
    result = runner.invoke(cmd, ['destroy'], obj=script_info)
    assert 1 == result.exit_code

    result = runner.invoke(cmd, ['destroy', '--yes-i-know'], obj=script_info)
    assert 0 == result.exit_code

    aliases = current_search_client.indices.get_alias()
    assert 0 == len(aliases)


def test_list(app):
    """Run listing of mappings."""
    suffix = '-abc'
    app.config['SEARCH_MAPPINGS'] = ['records']
    search = app.extensions['invenio-search']
    search._current_suffix = suffix
    search.register_mappings('authors', 'mock_module.mappings')
    search.register_mappings('records', 'mock_module.mappings')

    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda info: app)

    result = runner.invoke(cmd, ['list', '--only-aliases'], obj=script_info)
    # Turn cli outputted str presentation of Python list into a list
    assert set(ast.literal_eval(result.output)) == {'records', 'authors'}

    result = runner.invoke(cmd, ['list'], obj=script_info)
    assert result.output == (
        u"├──authors\n"
        u"│  └──authors-authors-v1.0.0\n"
        u"└──records *\n"
        u"   ├──records-authorities\n"
        u"   │  └──records-authorities-authority-v1.0.0\n"
        u"   ├──records-bibliographic\n"
        u"   │  └──records-bibliographic-bibliographic-v1.0.0\n"
        u"   └──records-default-v1.0.0\n\n"
    )


def test_check(app):
    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda info: app)

    result = runner.invoke(cmd, ['check'], obj=script_info)
    assert result.exit_code == 0

    with patch('invenio_search.cli.ES_VERSION',
               return_value=(ES_VERSION[0] + 1, 0, 0)):
        result = runner.invoke(cmd, ['check'], obj=script_info)
        assert result.exit_code != 0


def test_create_put_and_delete(app):
    runner = CliRunner()
    script_info = ScriptInfo(create_app=lambda info: app)
    name = 'test-index-name'

    result = runner.invoke(cmd, [
        'create', '--verbose', name,
        '--body', './tests/mock_module/mappings/authors/authors-v1.0.0.json',
    ], obj=script_info)
    assert result.exit_code == 0
    assert name in list(current_search_client.indices.get('*').keys())

    doc_type = '_doc' if ES_VERSION[0] > 5 else 'recid'
    result = runner.invoke(cmd, [
        'put', name, doc_type,
        '--verbose', '--identifier', 1,
        '--body', './tests/mock_module/mappings/authors/authors-v1.0.0.json',
    ], obj=script_info)
    assert result.exit_code == 0
    current_search_client.get(index=name, doc_type=doc_type, id=1)
    with pytest.raises(NotFoundError):
        current_search_client.get(index=name, doc_type=doc_type, id=2)

    result = runner.invoke(cmd, [
        'delete', '--verbose', '--yes-i-know', '--force', name,
    ], obj=script_info)
    assert result.exit_code == 0
    assert name not in list(current_search_client.indices.get('*').keys())



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utility functions for search engine."""

import os
import time
import warnings

import six
from elasticsearch import VERSION as ES_VERSION
from flask import current_app

from .proxies import current_search, current_search_client


def timestamp_suffix():
    """Generate a suffix based on the current time."""
    return '-' + str(int(time.time()))


def prefix_index(index, prefix=None, app=None):
    """Prefixes the given index if needed.

    :param index: Name of the index to prefix.
    :param prefix: Force a prefix.
    :param app: Flask app to get the prefix config from.
    :returns: A string with the new index name prefixed if needed.
    """
    app = app or current_app
    index_prefix = (prefix if prefix is not None
                    else (app.config.get('SEARCH_INDEX_PREFIX')) or '')
    return index_prefix + index


def suffix_index(index, suffix=None, app=None):
    """Suffixes the given index.

    :param index: Name of the index to prefix.
    :param suffix: The suffix to append to the index name.
    :param app: Flask app to get the "invenio-search" extension from.
    :returns: A string with the new index name suffixed.
    """
    search_ext = app.extensions['invenio-search'] if app else current_search
    suffix = suffix if suffix is not None else search_ext.current_suffix
    return index + suffix


def build_index_from_parts(*parts):
    """Build an index name from parts.

    :param parts: String values that will be joined by dashes ("-").
    """
    return '-'.join([part for part in parts if part])


def build_alias_name(index, prefix=None, app=None):
    """Build an alias name.

    :param index: Name of the index.
    :param prefix: The prefix to prepend to the index name.
    """
    return build_index_name(index, prefix=prefix, suffix='', app=app)


def build_index_name(index, prefix=None, suffix=None, app=None):
    """Build an index name.

    :param index: Name of the index.
    :param prefix: The prefix to prepend to the index name.
    :param suffix: The suffix to append to the index name.
    :param app: Flask app passed to ``prefix_index`` and ``suffix_index``.
    """
    if not isinstance(index, six.string_types):
        index = build_index_from_parts(*index)
    index = prefix_index(index, prefix=prefix, app=app)
    index = suffix_index(index, suffix=suffix, app=app)
    return index


def schema_to_index(schema, index_names=None):
    """Get index/doc_type given a schema URL.

    :param schema: The schema name
    :param index_names: A list of index name.
    :returns: A tuple containing (index, doc_type).
    """
    warnings.warn(
        '"invenio_search.utils.schema_to_index" will be moved to '
        'invenio-indexer',
        DeprecationWarning
    )
    parts = schema.split('/')
    doc_type, ext = os.path.splitext(parts[-1])
    parts[-1] = doc_type
    if ES_VERSION[0] >= 7:
        doc_type = '_doc'

    if ext not in {'.json', }:
        return (None, None)

    if index_names is None:
        index = build_index_from_parts(*parts)
        return index, doc_type

    for start in range(len(parts)):
        name = build_index_from_parts(*parts[start:])
        if name in index_names:
            return name, doc_type

    return (None, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for information retrieval."""

import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'check-manifest>=0.35',
    'coverage>=4.0',
    'invenio-db[versioning]>=1.0.0',
    'isort>=4.2.15',
    'mock>=1.3.0',
    'pydocstyle>=1.0.0',
    'pytest-cov>=1.8.0',
    'pytest-random-order>=0.5.4',
    "pytest-pep8>=1.0.6",
    'pytest>=3.8.1,<4',
]

extras_require = {
    'docs': [
        'Sphinx>=1.8.4,<3',
        'invenio-accounts>=1.0.0',
    ],
    # Elasticsearch version
    'elasticsearch2': [
        'elasticsearch>=2.0.0,<3.0.0',
        'elasticsearch-dsl>=2.0.0,<3.0.0',
    ],
    'elasticsearch5': [
        'elasticsearch>=5.0.0,<6.0.0',
        'elasticsearch-dsl>=5.1.0,<6.0.0',
    ],
    'elasticsearch6': [
        'elasticsearch>=6.0.0,<7.0.0',
        'elasticsearch-dsl>=6.0.0,<6.2.0',
    ],
    'elasticsearch7': [
        'elasticsearch>=7.0.0,<8.0.0',
        'elasticsearch-dsl>=7.0.0,<8.0.0',
    ],
    'tests': tests_require,
}

extras_require['all'] = []
for name, reqs in extras_require.items():
    if name[0] == ':' or name in (
            'elasticsearch2', 'elasticsearch5', 'elasticsearch6',
            'elasticsearch7'):
        continue
    extras_require['all'].extend(reqs)


setup_requires = [
    'pytest-runner>=2.6.2',
]

install_requires = [
    'invenio-base>=1.2.2',
]

packages = find_packages()

# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('invenio_search', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='invenio-search',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='invenio search',
    license='MIT',
    author='CERN',
    author_email='info@inveniosoftware.org',
    url='https://github.com/inveniosoftware/invenio-search',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={
        'invenio_base.api_apps': [
            'invenio_search = invenio_search:InvenioSearch',
        ],
        'invenio_base.apps': [
            'invenio_search = invenio_search:InvenioSearch',
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Development Status :: 5 - Production/Stable',
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Click command-line interface for managing search indexes."""

from __future__ import absolute_import, print_function

import json
import sys
from functools import wraps

import click
from elasticsearch import VERSION as ES_VERSION
from flask.cli import with_appcontext

from .proxies import current_search, current_search_client


def abort_if_false(ctx, param, value):
    """Abort command if value is False."""
    if not value:
        ctx.abort()


def es_version_check(f):
    """Decorator to check Elasticsearch version."""
    @wraps(f)
    def inner(*args, **kwargs):
        cluster_ver = current_search.cluster_version[0]
        client_ver = ES_VERSION[0]
        if cluster_ver != client_ver:
            raise click.ClickException(
                'Elasticsearch version mismatch. Invenio was installed with '
                'Elasticsearch v{client_ver}.x support, but the cluster runs '
                'Elasticsearch v{cluster_ver}.x.'.format(
                    client_ver=client_ver,
                    cluster_ver=cluster_ver,
                ))
        return f(*args, **kwargs)
    return inner


#
# Index management commands
#
@click.group()
def index():
    """Manage search indices."""


@index.command()
@with_appcontext
@es_version_check
def check():
    """Check Elasticsearch version."""
    click.secho('Checks passed', fg='green')


@index.command()
@click.option('--force', is_flag=True, default=False)
@with_appcontext
@es_version_check
def init(force):
    """Initialize registered aliases and mappings."""
    click.secho('Creating indexes...', fg='green', bold=True, file=sys.stderr)
    with click.progressbar(
            current_search.create(ignore=[400] if force else None),
            length=len(current_search.mappings)) as bar:
        for name, response in bar:
            bar.label = name
    click.secho('Putting templates...', fg='green', bold=True, file=sys.stderr)
    with click.progressbar(
            current_search.put_templates(ignore=[400] if force else None),
            length=len(current_search.templates)) as bar:
        for response in bar:
            bar.label = response


@index.command()
@click.option('--yes-i-know', is_flag=True, callback=abort_if_false,
              expose_value=False,
              prompt='Do you know that you are going to destroy all indexes?')
@click.option('--force', is_flag=True, default=False)
@with_appcontext
@es_version_check
def destroy(force):
    """Destroy all indexes."""
    click.secho('Destroying indexes...', fg='red', bold=True, file=sys.stderr)
    with click.progressbar(
            current_search.delete(ignore=[400, 404] if force else None),
            length=len(current_search.mappings)) as bar:
        for name, response in bar:
            bar.label = name


@index.command()
@click.argument('index_name')
@click.option('-b', '--body', type=click.File('r'), default=sys.stdin)
@click.option('--force', is_flag=True, default=False)
@click.option('--verbose', is_flag=True, default=False)
@with_appcontext
@es_version_check
def create(index_name, body, force, verbose):
    """Create a new index."""
    result = current_search_client.indices.create(
        index=index_name,
        body=json.load(body),
        ignore=[400] if force else None,
    )
    if verbose:
        click.echo(json.dumps(result))


@index.command('list')
@click.option('-a', '--only-active', is_flag=True, default=False)
@click.option('--only-aliases', is_flag=True, default=False)
@click.option('--verbose', is_flag=True, default=False)
@with_appcontext
def list_cmd(only_active, only_aliases, verbose):
    """List indices."""
    def _tree_print(d, rec_list=None, verbose=False, indent=2):
        # Note that on every recursion rec_list is copied,
        # which might not be very effective for very deep dictionaries.
        rec_list = rec_list or []
        for idx, key in enumerate(sorted(d)):
            line = (['│' + ' ' * indent
                     if i == 1 else ' ' * (indent+1) for i in rec_list])
            line.append('└──' if len(d)-1 == idx else '├──')
            click.echo(''.join(line), nl=False)
            if isinstance(d[key], dict):
                click.echo(key)
                new_rec_list = rec_list + [0 if len(d)-1 == idx else 1]
                _tree_print(d[key], new_rec_list, verbose)
            else:
                leaf_txt = '{} -> {}'.format(key, d[key]) if verbose else key
                click.echo(leaf_txt)

    aliases = (current_search.active_aliases
               if only_active else current_search.aliases)
    active_aliases = current_search.active_aliases

    if only_aliases:
        click.echo(json.dumps(list((aliases.keys())), indent=4))
    else:
        # Mark active indices for printout
        aliases = {(k + (' *' if k in active_aliases else '')): v
                   for k, v in aliases.items()}
        click.echo(_tree_print(aliases, verbose=verbose))


@index.command()
@click.argument('index_name')
@click.option('--force', is_flag=True, default=False)
@click.option('--verbose', is_flag=True, default=False)
@click.option('--yes-i-know', is_flag=True, callback=abort_if_false,
              expose_value=False,
              prompt='Do you know that you are going to delete the index?')
@with_appcontext
@es_version_check
def delete(index_name, force, verbose):
    """Delete index by its name."""
    result = current_search_client.indices.delete(
        index=index_name,
        ignore=[400, 404] if force else None,
    )
    if verbose:
        click.echo(json.dumps(result))


@index.command()
@click.argument('index_name')
@click.argument('doc_type')
@click.option('-i', '--identifier', default=None)
@click.option('-b', '--body', type=click.File('r'), default=sys.stdin)
@click.option('--force', is_flag=True, default=False)
@click.option('--verbose', is_flag=True, default=False)
@with_appcontext
@es_version_check
def put(index_name, doc_type, identifier, body, force, verbose):
    """Index input data."""
    result = current_search_client.index(
        index=index_name,
        doc_type=doc_type or index_name,
        id=identifier,
        body=json.load(body),
        op_type='index' if force or identifier is None else 'create',
    )
    if verbose:
        click.echo(json.dumps(result))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

from collections import defaultdict

import pytest
from elasticsearch import VERSION as ES_VERSION
from flask import Flask
from mock import patch

from invenio_search import InvenioSearch, current_search, current_search_client
from invenio_search.errors import IndexAlreadyExistsError


def test_version():
    """Test version import."""
    from invenio_search import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = InvenioSearch(app)
    assert 'invenio-search' in app.extensions

    app = Flask('testapp')
    ext = InvenioSearch()
    assert 'invenio-search' not in app.extensions
    ext.init_app(app)
    assert 'invenio-search' in app.extensions


def test_client_config():
    """Test Elasticsearch client configuration."""
    app = Flask('testapp')
    app.config['SEARCH_CLIENT_CONFIG'] = {'timeout': 30, 'foo': 'bar'}
    with patch('elasticsearch.Elasticsearch.__init__') as mock_es_init:
        mock_es_init.return_value = None
        ext = InvenioSearch(app)
        es_client = ext.client  # trigger client initialization
        mock_es_init.assert_called_once_with(
            hosts=None, timeout=30, foo='bar')


def test_flush_and_refresh(app):
    """Test flush and refresh."""
    search = app.extensions['invenio-search']
    search.flush_and_refresh('_all')


def test_client_reference():
    """Test client reference."""
    client1 = {'name': 'client1'}
    client2 = {'name': 'client2'}

    app1 = Flask('testapp1')
    app2 = Flask('testapp2')

    ext = InvenioSearch()
    assert 'invenio-search' not in app1.extensions
    assert 'invenio-search' not in app2.extensions

    ext.init_app(app1, client=client1)
    assert 'invenio-search' in app1.extensions

    ext.init_app(app2, client=client2)
    assert 'invenio-search' in app2.extensions

    with app1.app_context():
        assert current_search_client == client1

    with app2.app_context():
        assert current_search_client == client2


def test_default_client(app):
    """Test default client."""
    current_search_client.cluster.health(
        wait_for_status='yellow', request_timeout=1
    )


def test_load_entry_point_group(template_entrypoints):
    """Test entry point loading."""
    app = Flask('testapp')
    ext = InvenioSearch(app)
    ep_group = 'test'

    def mock_entry_points_mappings(group=None):
        assert group == ep_group

        class ep(object):

            name = 'records'
            module_name = 'mock_module.mappings'

        yield ep

    assert len(ext.mappings) == 0
    with patch('invenio_search.ext.iter_entry_points',
               mock_entry_points_mappings):
        ext.load_entry_point_group_mappings(
            entry_point_group_mappings=ep_group)
    assert len(ext.mappings) == 3
    # Check that mappings are loaded from the correct folder depending on the
    # ES version.
    if ES_VERSION[0] > 2:
        mappings_dir = 'mock_module/mappings/v{}/records'.format(ES_VERSION[0])
        assert all(mappings_dir in path for path in ext.mappings.values())

    with patch('invenio_search.ext.iter_entry_points',
               return_value=template_entrypoints('invenio_search.templates')):
        if ES_VERSION[0] == 2:
            assert set(ext.templates.keys()) == \
                {'subdirectory-file-download-v1', 'record-view-v1'}
        elif ES_VERSION[0] == 5:
            assert set(ext.templates.keys()) == \
                {'record-view-v{}'.format(ES_VERSION[0])}


@pytest.mark.parametrize(('aliases_config', 'expected_aliases'), [
    (None, ['records', 'authors']),
    (['records', 'authors'], ['records', 'authors']),
    (['authors'], ['authors']),
    ([], []),
    (['does_not_exist'], []),
])
def test_whitelisted_aliases(app, aliases_config, expected_aliases):
    """Test functionality of active aliases configuration variable."""
    all_aliases = dict(
        authors=['authors', 'authors-authors-v1.0.0'],
        records=[
            'records',
            'records-default-v1.0.0',
            'records-authorities',
            'records-authorities-authority-v1.0.0',
            'records-bibliographic',
            'records-bibliographic-bibliographic-v1.0.0',
        ]
    )

    orig = app.config['SEARCH_MAPPINGS']

    search = app.extensions['invenio-search']
    search.register_mappings('records', 'mock_module.mappings')
    search.register_mappings('authors', 'mock_module.mappings')

    app.config.update(SEARCH_MAPPINGS=aliases_config)

    current_search_client.indices.delete_alias('_all', '_all', ignore=[400,
                                                                       404])
    current_search_client.indices.delete('*')
    list(current_search.create(ignore=None))

    aliases = current_search_client.indices.get_alias()
    if expected_aliases == []:
        assert 0 == len(aliases)
    else:
        for expected_alias in expected_aliases:
            all_expected = all_aliases[expected_alias]
            assert current_search_client.indices.exists(all_expected)

    app.config['SEARCH_MAPPINGS'] = orig


@pytest.mark.parametrize(
    ('suffix', 'create_index', 'create_alias', 'expected'), [
        ('', 'authors-authors-v1.0.0', None, []),
        ('', None, 'authors-authors-v1.0.0', []),
        ('-abc', 'authors-authors-v1.0.0', None, []),
        ('-test', 'authors-authors-not-exist', None, [
            'authors-authors-v1.0.0-test',
            'authors-authors-v1.0.0',
            'authors',
        ]),
    ]
)
def test_creating_alias_existing_index(app, suffix, create_index, create_alias,
                                       expected):
    """Test creating new alias and index where there already exists one."""
    search = app.extensions['invenio-search']
    search.register_mappings('authors', 'mock_module.mappings')
    search._current_suffix = suffix
    current_search_client.indices.delete_alias('_all', '_all', ignore=[400,
                                                                       404])
    current_search_client.indices.delete('*')
    new_indexes = []
    if create_index:
        current_search_client.indices.create(index=create_index)
        new_indexes.append(create_index)
    if create_alias:
        write_alias_index = '{}-suffix'.format(create_alias)
        current_search_client.indices.create(
            index=write_alias_index
        )
        new_indexes.append(write_alias_index)
        current_search_client.indices.put_alias(
            index=write_alias_index,
            name=create_alias,
        )
    if expected:
        results = list(current_search.create(ignore=None))
        assert len(results) == len(expected)
        for result in results:
            assert result[0] in expected
        indices = current_search_client.indices.get('*')
        index_names = list(indices.keys())
        alias_names = []
        for index in index_names:
            alias_names.extend(list(indices[index]['aliases'].keys()))
        for index, _ in results:
            if index.endswith(suffix):
                assert sorted(index_names) == sorted([index] + new_indexes)
            else:
                assert index in alias_names
    else:
        with pytest.raises(Exception):
            results = list(current_search.create(ignore=None))
        indices = current_search_client.indices.get('*')
        index_names = list(indices.keys())
        assert index_names == new_indexes
        if create_index:
            assert len(indices[create_index]['aliases']) == 0


@pytest.mark.parametrize(('aliases_config', 'prefix', 'expected_aliases'), [
    (['authors'], 'test-', ['authors']),
    (['authors', 'records'], 'dev-', ['records', 'authors']),
])
def test_prefix_search_mappings(app, aliases_config, prefix, expected_aliases):
    """Test that indices are created when prefix & search mappings are set."""
    app.config.update(
        SEARCH_MAPPINGS=aliases_config,
        SEARCH_INDEX_PREFIX=prefix,
    )

    search = app.extensions['invenio-search']
    search.register_mappings('records', 'mock_module.mappings')
    search.register_mappings('authors', 'mock_module.mappings')

    active_aliases = search.active_aliases

    assert len(active_aliases) == len(expected_aliases)
    for expected_alias in expected_aliases:
        assert expected_alias in active_aliases


def _test_prefix_indices(app, prefix_value):
    """Assert that each index name contains the prefix."""
    app.config['SEARCH_INDEX_PREFIX'] = prefix_value
    suffix = '-abc'
    prefix = prefix_value or ''
    search = app.extensions['invenio-search']
    search._current_suffix = suffix
    search.register_mappings('records', 'mock_module.mappings')

    # clean-up in case something failed previously
    current_search_client.indices.delete('*')
    # create indices and test
    list(search.create())
    es_indices = current_search_client.indices.get_alias()

    def _f(name):  # formatting helper
        return name.format(p=prefix, s=suffix)

    assert set(es_indices.keys()) == {
        _f('{p}records-authorities-authority-v1.0.0{s}'),
        _f('{p}records-bibliographic-bibliographic-v1.0.0{s}'),
        _f('{p}records-default-v1.0.0{s}'),
    }
    # Build set of aliases
    es_aliases = defaultdict(set)
    for index, info in es_indices.items():
        for alias in info.get('aliases', {}):
            es_aliases[alias].add(index)

    auth_idx = {_f('{p}records-authorities-authority-v1.0.0{s}')}
    bib_idx = {_f('{p}records-bibliographic-bibliographic-v1.0.0{s}')}
    default_idx = {_f('{p}records-default-v1.0.0{s}')}
    all_indices = auth_idx | bib_idx | default_idx
    assert es_aliases == {
        _f('{p}records-authorities-authority-v1.0.0'): auth_idx,
        _f('{p}records-bibliographic-bibliographic-v1.0.0'): bib_idx,
        _f('{p}records-default-v1.0.0'): default_idx,
        _f('{p}records-authorities'): auth_idx,
        _f('{p}records-bibliographic'): bib_idx,
        _f('{p}records'): all_indices,
    }
    # clean-up
    current_search_client.indices.delete('*')


def test_indices_prefix_empty_value(app):
    """Test indices creation with prefix value empty string."""
    prefix_value = ''
    _test_prefix_indices(app, prefix_value)


def test_indices_prefix_none_value(app):
    """Test indices creation with a prefix value None."""
    prefix_value = None
    _test_prefix_indices(app, prefix_value)


def test_indices_prefix_some_value(app):
    """Test indices creation with a prefix value `myprefix-`."""
    prefix_value = 'myprefix-'
    _test_prefix_indices(app, prefix_value)


def _test_prefix_templates(app, prefix_value, template_entrypoints):
    """Assert that templates take into account the prefix name."""
    def _contains_prefix(prefix_value, string_or_list):
        """Return True if the prefix value is in the input list or string."""
        if isinstance(string_or_list, list):
            return all([True for match in string_or_list
                        if prefix_value in match])
        else:
            return prefix_value in string_or_list

    def _test_prefix_replaced_in_body(name, prefix_value, tpl_key):
        """Test that the prefix is replaced in the body when defined."""
        if prefix_value:
            tpl = current_search_client.indices.get_template(name)
            assert name in tpl
            assert _contains_prefix(prefix_value, tpl[name][tpl_key])

    app.config['SEARCH_INDEX_PREFIX'] = prefix_value
    search = app.extensions['invenio-search']
    with patch('invenio_search.ext.iter_entry_points',
               return_value=template_entrypoints('invenio_search.templates')):
        # clean-up in case something failed previously
        current_search_client.indices.delete_template('*')
        # create templates
        list(search.put_templates())

        if ES_VERSION[0] == 2:
            assert len(search.templates.keys()) == 2
            name = 'record-view-v1'
            prefixed = (prefix_value or '') + name
            assert name in search.templates
            assert current_search_client.indices.exists_template(prefixed)
            _test_prefix_replaced_in_body(prefixed, prefix_value, 'template')
            name = 'subdirectory-file-download-v1'
            prefixed = (prefix_value or '') + name
            assert name in search.templates
            assert current_search_client.indices.exists_template(prefixed)
            _test_prefix_replaced_in_body(prefixed, prefix_value, 'template')
        elif ES_VERSION[0] == 5:
            assert len(search.templates.keys()) == 1
            name = 'record-view-v{0}'.format(ES_VERSION[0])
            prefixed = (prefix_value or '') + name
            assert name in search.templates
            assert current_search_client.indices.exists_template(prefixed)
            _test_prefix_replaced_in_body(prefixed, prefix_value, 'template')
        else:
            assert len(search.templates.keys()) == 1
            name = 'record-view-v{0}'.format(ES_VERSION[0])
            prefixed = (prefix_value or '') + name
            assert name in search.templates
            assert current_search_client.indices.exists_template(prefixed)
            _test_prefix_replaced_in_body(
                prefixed, prefix_value, 'index_patterns')

        # clean-up
        current_search_client.indices.delete_template('*')


def test_templates_prefix_empty_value(app, template_entrypoints):
    """Test templates creation with prefix value empty string."""
    prefix_value = ''
    _test_prefix_templates(app, prefix_value, template_entrypoints)


def test_templates_prefix_none_value(app, template_entrypoints):
    """Test templates creation with a prefix value None."""
    prefix_value = None
    _test_prefix_templates(app, prefix_value, template_entrypoints)


def test_templates_prefix_some_value(app, template_entrypoints):
    """Test templates creation with a prefix value `myprefix-`."""
    prefix_value = 'myprefix-'
    _test_prefix_templates(app, prefix_value, template_entrypoints)


def test_current_suffix(app):
    """Test generating current suffix."""
    search = app.extensions['invenio-search']
    suffix = search.current_suffix
    assert suffix == search.current_suffix


def test_not_dry_run_and_index_exists(app):
    """Test create_index and no dry run when index exists."""
    current_search_client.indices.delete('*')
    current_search_client.indices.create(
        index='records-default-v1.0.0',
        body=""
    )
    search = app.extensions['invenio-search']
    search.register_mappings('records', 'mock_module.mappings')
    with pytest.raises(IndexAlreadyExistsError):
        list(search.create())


def test_create_selected_indexes(app):
    search = app.extensions['invenio-search']
    current_search_client.indices.delete('*')
    search.register_mappings('authors', 'mock_module.mappings')
    search.register_mappings('records', 'mock_module.mappings')
    list(search.create(index_list=[
        'records-bibliographic-bibliographic-v1.0.0'
    ]))
    assert search.client.indices.exists(
        'records-bibliographic-bibliographic-v1.0.0') is True
    assert search.client.indices.exists_alias(
        'records-bibliographic-bibliographic-v1.0.0',
        'records,'
        'records-bibliographic,'
        'records-bibliographic-bibliographic-v1.0.0') is True
    assert search.client.indices.exists('records-default-v1.0.0') is False
    assert search.client.indices.exists('authors-authors-v1.0.0') is False
    assert search.client.indices.exists(
        'records-authorities-authority-v1.0.0') is False


def test_delete_selected_indexes(app):
    search = app.extensions['invenio-search']
    current_search_client.indices.delete('*')
    search.register_mappings('authors', 'mock_module.mappings')
    search.register_mappings('records', 'mock_module.mappings')
    list(search.create())
    list(search.delete(index_list=['authors-authors-v1.0.0']))
    assert search.client.indices.exists('authors-authors-v1.0.0') is False
    assert search.client.indices.exists('records-default-v1.0.0') is True
    assert search.client.indices.exists(
        'records-bibliographic-bibliographic-v1.0.0') is True


def test_create_when_indexes_already_exists_with_ignore_existing_true(app):
    search = app.extensions['invenio-search']
    current_search_client.indices.delete('*')
    search.register_mappings('authors', 'mock_module.mappings')
    search.register_mappings('records', 'mock_module.mappings')
    list(search.create(index_list=[
        'authors-authors-v1.0.0',
        'records-default-v1.0.0'
    ]))
    list(search.create(ignore_existing=True))
    assert search.client.indices.exists(
        'records-bibliographic-bibliographic-v1.0.0') is True
